import os
import asyncio
import json
import serial
from datetime import datetime
import traceback

from .settings import DEFAULT_MODEM_DEV, DEFAULT_4G_MODEM_DEV, WB_HARDWARE_CONF_FILE_PATH, PPP_PROVIDER_FILE_PATH
from .sys_utils import ShellCommand


PROVIDER_ERROR = 0
PROVIDER_NEW = 1
PROVIDER_OK = 2

DELAY = 15

# MODEM_ON = True
# MODEM_OFF = False

MODULE_4G = 'wbc-4g'

MODEM_DEV_FILEPATH = 'dev_filepath'
MODEM_TYPE = 'type'

MODEM_TYPE_DEFAULT = 'DEFAULT'
MODEM_TYPE_3G = '3G'
MODEM_TYPE_4G = '4G'

# MODEM_DEFAULT_SETTINGS = {
#     'dev_filepath': DEFAULT_MODEM_DEV
# }


class GSMModem(ShellCommand):
    def __init__(self, logger=None):
        super().__init__(logger)
        self._provider_filepath = PPP_PROVIDER_FILE_PATH

        # self._dev_filepath = DEFAULT_MODEM_DEV
        self._settings = {
            MODEM_DEV_FILEPATH: DEFAULT_MODEM_DEV,
            MODEM_TYPE: MODEM_TYPE_DEFAULT
        }
        self._is_4g = self._4g()

    def is_4g(self):
        return self._is_4g

    def _4g(self):
        if self._4g_check():
            self._settings[MODEM_DEV_FILEPATH] = DEFAULT_4G_MODEM_DEV
            self._settings[MODEM_TYPE] = MODEM_TYPE_4G
            return True
        return False

    def _4g_check(self):
        if os.path.isfile(WB_HARDWARE_CONF_FILE_PATH):
            try:
                with open(WB_HARDWARE_CONF_FILE_PATH) as f:
                    json_conf = json.load(f)
                    slots = json_conf.get('slots')
                    if (slots is None) or (not slots):
                        return False
                    for slot in slots:
                        module = slot.get('module')
                        if module is None:
                            continue
                        if module == MODULE_4G:
                            return True
            except:
                self.logger.error("Can't read WB conf json-file [{0}]:\n{1}".format(WB_HARDWARE_CONF_FILE_PATH,
                                                                                    traceback.format_exc()))
        return False

    def settings(self):
        return self._settings

    # $ wb-gsm on - включение устройства
    async def on(self):
        await self.command_exec("/usr/bin/wb-gsm on")
        # self.state = MODEM_ON
        await asyncio.sleep(DELAY)

    # $ wb-gsm off - выключение устройства
    async def _off(self):
        await self.command_exec("/usr/bin/wb-gsm off")
        # self.state = MODEM_OFF
        await asyncio.sleep(DELAY)

    # $ wb-gsm reset - перезагрузка включенного устройства
    async def reset(self):
        await self.on()
        await self.command_exec("/usr/bin/wb-gsm reset")
        # self.state = MODEM_ON
        await asyncio.sleep(DELAY)

    # $ wb-gsm restart_if_broken - включение или перезагрузка, если модем не отвечает
    async def restart(self):
        await self._off()
        await self.command_exec("/usr/bin/wb-gsm restart_if_broken")
        await asyncio.sleep(DELAY)

    async def check(self):
        if not self._is_4g:
            modem3g_dev_template = '/dev/ttyACM{0}'
            for i in range(8):
                dev_tmp = modem3g_dev_template.format(i)
                if await self._is_modem3g(dev_tmp):
                    self._settings[MODEM_DEV_FILEPATH] = dev_tmp
                    self._settings[MODEM_TYPE] = MODEM_TYPE_3G
                    break

        provider_status = self._check_modem_provider()
        if provider_status == PROVIDER_ERROR:
            self.logger.error("GSM modem provider error...")
            # await self.off()
        elif provider_status == PROVIDER_NEW:
            self.logger.info("GSM modem update provider... reset")
            await self.reset()
        else:
            # await self.on()
            pass
        # return self.state

    async def _is_modem3g(self, dev_filepath):
        result = False
        if os.path.exists(dev_filepath):
            cmd = "udevadm info {0} --query=property".format(dev_filepath)
            res, out, err = await self.command_exec(cmd, log=False)
            if res != 0:
                return result
            try:
                id_model = ''
                id_vendor = ''
                for line in [s.strip() for s in out.splitlines()]:
                    # print("line [{0}]".format(line))
                    if not line:
                        continue
                    tmp = line.split('=')
                    if len(tmp) == 2:
                        key = tmp[0].strip()
                        value = tmp[1].strip()
                        if key == 'ID_MODEL':
                            id_model = value
                        elif key == 'ID_VENDOR':
                            id_vendor = value
                    # print("id_model=[{0}]; id_vendor=[{1}]".format(id_model, id_vendor))
                if (id_model == "SIMCOM_PRODUCT") and (id_vendor == "SIMCOM_VENDOR"):
                    result = True
            except:
                result = False
        return result

    def _check_modem_provider(self):
        if not os.path.isfile(self._provider_filepath):
            return PROVIDER_ERROR

        dev_filepath_template = "/dev/tty"
        # result = PROVIDER_NEW
        new_content = []
        once_flag = True
        with open(self._provider_filepath) as f:
            for line in f:
                line_tmp = line.strip()
                if once_flag and ((not line_tmp) or line_tmp.startswith("#")):
                    continue

                if once_flag and line_tmp.startswith(dev_filepath_template):
                    if line_tmp == self._settings[MODEM_DEV_FILEPATH]:
                        return PROVIDER_OK
                        # break
                    else:
                        line = "# [IQnergy, {0}]\n#{1}\n".format(datetime.now(), line)
                        line += "# New modem dev:\n{0}\n".format(self._settings[MODEM_DEV_FILEPATH])
                    once_flag = False
                new_content.append(line)

        # if result == PROVIDER_NEW:
        with open(self._provider_filepath, 'w') as file:
            file.writelines(new_content)
        return PROVIDER_NEW  # result

    async def _is_wb6(self):
        result = False
        cmd = "lscpu | grep Architecture"
        res, out, err = await self.command_exec(cmd, log=False)
        if res != 0:
            return result

        try:
            architecture = out.split()[1]
            if architecture == 'armv7l':
                result = True
        except Exception:
            pass
        return result

    async def change_sim_slot(self):
        if not await self._is_wb6():
            return

        wb_filepath = "/usr/lib/wb-utils/wb-gsm-common.sh"
        if not os.path.isfile(wb_filepath):
            return

        target = "gpio_set_value $WB_GPIO_GSM_SIMSELECT"
        # with open(wb_filepath) as f:
        #     content = f.readlines()

        # content = [x.strip() for x in content]
        new_content = []
        sim_slot = 0
        once_flag = True

        # for line in content:
        #     target_idx = line.find(target)
        #     new_line = line
        #     line_tmp = line.strip()
        #     if line_tmp and (not line_tmp.startswith("#")) and (target_idx >= 0):
        #         if once_flag:
        #             try:
        #                 current_sim_slot = int(line_tmp.split()[-1])
        #                 sim_slot = 0 if current_sim_slot > 0 else 1
        #             except Exception:
        #                 pass
        #             once_flag = False
        #         new_line = "{0}{1} {2}\n".format(line[:target_idx], target, sim_slot)
        #     new_content.append(new_line)

        with open(wb_filepath) as f:
            for line in f:
                target_idx = line.find(target)
                new_line = line
                line_tmp = line.strip()
                if line_tmp and (not line_tmp.startswith("#")) and (target_idx >= 0):
                    if once_flag:
                        try:
                            current_sim_slot = int(line_tmp.split()[-1])
                            sim_slot = 0 if current_sim_slot > 0 else 1
                        except Exception:
                            pass
                        once_flag = False
                    new_line = "{0}{1} {2}\n".format(line[:target_idx], target, sim_slot)
                new_content.append(new_line)

        with open(wb_filepath, 'w') as file:
            file.writelines(new_content)

        self.logger.info("NEW SIM SLOT {0}".format(sim_slot))

    async def sim_slot(self):
        if not await self._is_wb6():
            return

        wb_filepath = "/usr/lib/wb-utils/wb-gsm-common.sh"
        if not os.path.isfile(wb_filepath):
            return

        target = "gpio_set_value $WB_GPIO_GSM_SIMSELECT"
        # with open(wb_filepath) as f:
        #     content = f.readlines()

        # content = [x.strip() for x in content]

        with open(wb_filepath) as f:
            for line in f:
                line_tmp = line.strip()
                if line_tmp and (not line_tmp.startswith("#")) and (target in line_tmp):
                    try:
                        current_sim_slot = int(line_tmp.split()[-1])
                        sim_slot = 1 if current_sim_slot > 0 else 0
                        return sim_slot
                    except Exception:
                        pass
                    return

    async def quality(self):
        if self._settings[MODEM_DEV_FILEPATH] == DEFAULT_MODEM_DEV:
            return
        result = {}
        signal_quality = self._signal_quality()
        if (signal_quality is not None) and signal_quality:
            result['signal quality'] = signal_quality
        sim_slot = await self.sim_slot()
        if sim_slot is not None:
            result['sim slot'] = sim_slot
        return result

    def at_command_exec(self, cmd):
        try:
            ser = serial.Serial(port=DEFAULT_MODEM_DEV, baudrate=115200, bytesize=8, parity='N', stopbits=1, timeout=1)
        except:
            self.logger.error(traceback.format_exc())
            return None
        try:
            if not cmd.endswith('\r'):
                cmd += '\r'
            ser.write(cmd.encode())
            msg = ser.read(64).decode("utf-8")
        except Exception:
            self.logger.error(traceback.format_exc())
            msg = None
        finally:
            ser.close()
        return msg

    def _signal_quality(self):
        msg = self.at_command_exec('AT+CSQ')
        # print("_signal_quality = {0}".format(msg))
        if msg is None:
            self.logger.warning("Error result of 'AT+CSQ'")
            return

        tmp = msg.split()
        tmp_len = len(tmp)
        str_result = None
        if tmp_len > 1:
            for i in range(tmp_len-1):
                if tmp[i].startswith('+CSQ:'):
                    str_result = tmp[i+1]
                    break
        if (str_result is None) or (not str_result):
            self.logger.warning("Can't parse AT+CSQ output: [{0}]".format(msg))
            return
        return str_result

    # def usb0_initialization(self, reset=False):
    #     if reset:
    #         print("0. {0}".format(self.at_command_exec('AT&F')))  # modem reset
    #
    #     print("1. {0}".format(self.at_command_exec('AT+DIALMODE=0')))
    #     print("2. {0}".format(self.at_command_exec('AT+CGDCONT=1,"IP","xxx"')))
    #     print("3. {0}".format(self.at_command_exec('AT+CGCONTRDP')))
