import json
# import time
import traceback
import threading
import asyncio

import paho.mqtt.client as mqtt

from .settings import SETTINGS_MQTT_TOPIC, RESULT_MQTT_TOPIC, STAT_MQTT_TOPIC
from .settings import DEFAULT_SETTINGS  #, DEFAULT_TIMEOUT, STORAGE_MQTT_TOPIC
from .sys_utils import IQLog


def mqtt_method(function):
    def wrapper(self, *args, **kwargs):
        # print(">> \"{0}\"".format(msg.topic))
        try:
            function(self, *args, **kwargs)
        except:
            self.logger.critical(traceback.format_exc())
        # print("<< \"{0}\"".format(msg.topic))
    return wrapper


class MQTTClient(IQLog):
    def __init__(self, logger=None):
        super().__init__(logger)

        self.connected = False
        self._settings_mutex = threading.Lock()
        self._settings = None
        # self._settings_are_changed = True

        self._client = mqtt.Client()
        self._message_callbacks = [
            (SETTINGS_MQTT_TOPIC, self._on_settings)
        ]

        self._client.on_connect = self._on_connect
        self._client.on_disconnect = self._on_disconnect

    def _message_callback_add(self):
        for topic, callback in self._message_callbacks:
            self._client.message_callback_add(topic, callback)

    @mqtt_method
    def _on_settings(self, client, userdata, msg):
        settings = self._process_json_payload(msg)
        # if settings is None:
        #     settings = {}
        with self._settings_mutex:
            self._settings = settings

    # def settings_are_changed(self, settings):
    #     with self._settings_mutex:
    #         if (settings is not None) and (self._settings is not None):
    #             return settings != self._settings
    #         if (settings is None) and (self._settings is None):
    #             return False
    #         # if ((settings is None) and (self._settings is not None)) or \
    #         #         ((settings is not None) and (self._settings is None)):
    #         #     return True
    #         return True

    def get_settings(self):
        with self._settings_mutex:
            return self._settings

    async def start(self):
        try:
            self._message_callback_add()

            # self._client.connect('localhost')
            # self._client.loop_forever()

            await self._start_connect()
            self._client.loop_start()

            while True:
                await asyncio.sleep(1)
        except:
            self.logger.critical(traceback.format_exc())

    async def _start_connect(self):
        try:
            self._client.connect('localhost')
        except:
            self.logger.error("{0}\nSLEEP 30 sec".format(traceback.format_exc()))
            await asyncio.sleep(30)
            await self._start_connect()

    @mqtt_method
    def _on_connect(self, client, userdata, flags, rc):
        txt = "Connection returned result: " + mqtt.connack_string(rc)
        if rc != mqtt.CONNACK_ACCEPTED:
            self.logger.critical(txt)
        else:
            self.connected = True
            self.logger.info(txt)
        client.subscribe([(topic, 1) for (topic, callback) in self._message_callbacks])

    @mqtt_method
    def _on_disconnect(self, client, userdata, rc):
        self.connected = False
        txt = "Disconnect returned result: " + mqtt.connack_string(rc)
        if rc != 0:
            self.logger.error("{0} [{1}]".format(txt, "Unexpected disconnection"))
            # client.reconnect()  # loop_start()/loop_forever() automatically handles reconnecting
        else:
            self.logger.warning(txt)  # disconnect by disconnect() method of client

    def _process_json_payload(self, msg):
        try:
            payload = json.loads(msg.payload.decode("utf-8"))
        except:
            payload = None
            self.logger.error(traceback.format_exc())
        return payload

    def _publish(self, topic, payload=None, qos=2, retain=True):
        try:
            # self.logger.info("start publish")
            publish_info = self._client.publish(topic, payload, qos, retain)
            # self.logger.info("publish ... ok")
            result = True if publish_info.rc == mqtt.MQTT_ERR_SUCCESS else False
        except:
            result = False
            self.logger.error(traceback.format_exc())
        if not result:
            err_txt = "Cannot send message with topic [{0}]: \"{1}\"".format(topic, payload)
            self.logger.error(err_txt)
        return result

    def _publish_as_json(self, topic, payload=None, qos=2, retain=True):
        try:
            # self.logger.info("get json_payload")
            json_payload = json.dumps(payload, separators=(',', ':'), sort_keys=True)
            # self.logger.info("json_payload ... ok")
        except:
            self.logger.error(traceback.format_exc())
            return False
        return self._publish(topic, json_payload, qos, retain)

    def publish_result(self, result):
        return self._publish_as_json(RESULT_MQTT_TOPIC, result)

    def publish_stat(self, stat):
        return self._publish_as_json(STAT_MQTT_TOPIC, stat)
