import os
import asyncio
import shutil
# import serial
import datetime
import traceback

from .sys_utils import ShellCommand
from .settings import LINUX_INTERFACES_FILE_PATH, LINUX_INTERFACES_BACKUP_FILE_PATH,\
    ETH0_INTERFACE_NAME, ETH1_INTERFACE_NAME, PPP_INTERFACE_NAME, USB0_INTERFACE_NAME, \
    PPP_PROVIDER, GSM_TIMEOUT, WLAN_INTERFACE_NAME, EMPTY_MODEM_SETTINGS

from .settings import DEFAULT_TIMEOUT

# DELAY = 5

# PUBLIC_DNS_LIST = ["1.1.1.1", "1.0.0.1",
#                    "84.200.69.80", "84.200.70.40",
#                    "8.8.8.8", "8.8.4.4",
#                    "208.67.222.222", "208.67.220.22"]
# PUBLIC_DNS = [dns.strip() for dns in PUBLIC_DNS_LIST]
PUBLIC_DNS = ["1.1.1.1", "1.0.0.1",
                   "84.200.69.80", "84.200.70.40",
                   "8.8.8.8", "8.8.4.4",
                   "208.67.222.222", "208.67.220.22"]


def check_dns():
    resolvconf_dir = "/etc/resolvconf/resolv.conf.d"
    resolv_conf_filepath = os.path.join(resolvconf_dir, "tail")
    if not os.path.isdir(resolvconf_dir):
        return False

    new_dns = list(PUBLIC_DNS)
    i = 0
    try:
        with open(resolv_conf_filepath) as file:
            for line in file:
                i += 1
                line = line.strip()
                if line and (not line.startswith("#")):
                    for dns in PUBLIC_DNS:
                        if dns in line:
                            try:
                                new_dns.remove(dns)
                            except ValueError:
                                pass
    except FileNotFoundError:
        pass

    if new_dns:
        nameserver_t = "nameserver {0}\n"
        lines = [nameserver_t.format(dns) for dns in new_dns]
        lines.insert(0, "# [IQnergy, {0}]\n".format(datetime.datetime.now()))
        if i > 0:
            lines.insert(0, "\n")
        with open(resolv_conf_filepath, 'a') as file:
            file.writelines(lines)
    return True


class NetUtils(ShellCommand):
    def __init__(self, modem=None, logger=None):
        super().__init__(logger)
        self._modem = modem

    async def print_routes(self):
        await self.command_exec("route -n")

    def check_dns(self):
        try:
            if not check_dns():
                self.logger.error("There is problem with [resolvconf]")
        except:
            self.logger.critical(traceback.format_exc())

    async def modem_on(self):
        if self._modem is not None:
            await self._modem.on()

    async def check_modem(self):
        if self._modem is not None:
            await self._modem.check()

    def is_4g_modem(self):
        if self._modem is not None:
            return self._modem.is_4g()
        return False

    def modem_settings(self):
        if self._modem is not None:
            return self._modem.settings()
        return EMPTY_MODEM_SETTINGS

    async def modem_quality(self):
        if self._modem is not None:
            return await self._modem.quality()

    @staticmethod
    def get_current_gateways():
        result = {}
        route_filepath = "/proc/net/route"
        try:
            with open(route_filepath) as f:
                headers = {}
                header_flag = True
                for line in f:
                    elems = line.split()
                    if header_flag:
                        i = 0
                        for elem in elems:
                            headers[elem.lower()] = i
                            i += 1
                        header_flag = False

                        for key_elem in ['iface', 'destination', 'flags', 'metric']:
                            if key_elem not in headers:
                                break
                        continue

                    if len(elems) < len(headers):
                        continue

                    # if fields[1] != '00000000' or not int(fields[3], 16) & 2:
                    if (elems[headers['destination']] == '00000000') or (int(elems[headers['flags']], 16) & 2):
                        iface = elems[headers['iface']]
                        metric = int(elems[headers['metric']])
                        metrics = result.get(iface)
                        if metrics is None:
                            metrics = []
                            result[iface] = metrics
                        metrics.append(metric)
        except Exception:
            pass
        return result

    # async def restart_networking_service(self):
    #     res, out, err = await self.command_exec("/etc/init.d/networking restart")
    #     # await asyncio.sleep(DELAY)
    #     return res

    async def delete_default_gateways(self, interface=None):
        await self.delete_default_gateways_template(interface,
                                                    lambda x, y: False if (x is not None) and (x != y) else True)

    async def delete_other_default_gateways(self, interface):
        await self.delete_default_gateways_template(interface,
                                                    lambda x, y: True if x != y else False)

    async def delete_default_gateways_template(self, input_interface, interface_filter=None):
        interfaces = {}

        gateway_interfaces = self.get_current_gateways()
        for interface_name, interface_metrics in gateway_interfaces.items():
            if (interface_filter is None) or \
                                ((interface_filter is not None) and
                                 (interface_filter(input_interface, interface_name))):
                interfaces[interface_name] = len(interface_metrics)

        for next_interface, num_gateway_routes in interfaces.items():
            for i in range(num_gateway_routes + 1):
                # print("{0}/{1}. {3} [{4}]: {2}".format(i+1, num_routes+1, next_interface, interface_filter, prefix))
                if await self._del_interface_gateway(next_interface) != 0:
                    # print("BREAK")
                    break

    async def _del_interface_gateway(self, interface_name):
        res, out, err = await self.command_exec("route del default dev {0}".format(interface_name))
        # await asyncio.sleep(DELAY)
        return res

    async def ping_interface(self, hostname, interface=None, count=1, timeout_sec=15):
        cmd = "timeout {2} ping {0} -c {1} -w {2}".format(hostname, count, timeout_sec)
        if interface is not None:
            cmd += " -I {0}".format(interface)
        cmd += " > /dev/null"

        res, out, err = await self.command_exec(cmd, log=False)
        if res != 0:
            return False
        return True

    async def check_internet(self, interface=None):
        test_hosts = ["my.iqnergy.com", "www.google.com"]
        n = 0
        for hostname in test_hosts:
            result_flag = await self.ping_interface(hostname, interface)
            if result_flag:
                n += 1

        if n > 0:
            return True
        return False

    async def internet_via_eth(self):
        for interface in [ETH0_INTERFACE_NAME, ETH1_INTERFACE_NAME]:
            if await self.check_internet(interface):
                return True
        return False

    async def check_interface(self, interface_name, current_metrics, target_metric):
        # recover_interface_flag = True
        if (len(current_metrics) == 1) and (current_metrics[0] == target_metric):
            if await self.check_internet(interface_name):
                # recover_interface_flag = False
                return True
            else:
                # delete_default_gateways(interface_name)
                self.logger.warning(" - NO INTERNET! Interface [{0}]".format(interface_name))
        # if recover_interface_flag:
        if await self.recover_interface(interface_name, target_metric):
            self.logger.info(" - RECOVER OK! Interface [{0}]".format(interface_name))
            return True
        else:
            self.logger.warning(" - RECOVER FAILED, NO INTERNET! Interface [{0}]".format(interface_name))
        return False

    async def _add_gateway(self, interface, metric):
        if interface == PPP_INTERFACE_NAME:
            await self.add_ppp_gateway(interface, metric=metric)
        else:
            await self.dhclient(interface, metric=metric)
        await asyncio.sleep(DEFAULT_TIMEOUT)

    async def recover_interface(self, interface, metric=0):
        await self.delete_default_gateways(interface)  # only if NO INTERNET at interface

        if skip_interface(interface):
            self.logger.warning(" - Physical link is down! Skip [{0}] interface".format(interface))
            return False

        self.logger.info(" - Stage 1 [{0}, {1}]".format(interface, metric))

        await self._add_gateway(interface, metric)

        # if interface == PPP_INTERFACE_NAME:
        #     await self.add_ppp_gateway(interface, metric=metric)
        # else:
        #     await self.dhclient(interface, metric=metric)

        if await self.check_internet(interface):
            return True

        self.logger.info(" - Stage 2 [{0}, {1}]".format(interface, metric))

        await self.ifdown(interface)
        await self.ifup(interface)

        await self._add_gateway(interface, metric)
        # if interface == PPP_INTERFACE_NAME:
        #     await self.add_ppp_gateway(interface, metric=metric)
        # else:
        #     await self.dhclient(interface, metric=metric)

        if await self.check_internet(interface):
            return True

        if ((interface == PPP_INTERFACE_NAME) or (interface == USB0_INTERFACE_NAME)) and (self._modem is not None):
            self.logger.info(" - Stage 3 (GSM reset)")
            await self.ifdown(interface)

            await self._modem.change_sim_slot()
            await self._modem.restart()

            await self.ifup(interface)

            await self._add_gateway(interface, metric)
            # await self.add_ppp_gateway(interface, metric=metric)

            if await self.check_internet(interface):
                return True

        await self.delete_default_gateways(interface)  # only if NO INTERNET at interface
        return False

    async def add_ppp_gateway(self, interface_name, metric=100):
        cmd = "route add default dev {0} metric {1}".format(interface_name, metric)
        return await self.command_exec(cmd)

    async def dhclient(self, interface, metric=100):
        # cmd = "dhclient -r && dhclient -1 {0}".format(interface)
        # TODO timeout!
        cmd = "/sbin/dhclient -1 -e IF_METRIC={1} {0}".format(interface, metric)
        # # cmd = "dhclient -1 {0}".format(interface)
        # result = command_exec(cmd)
        # time.sleep(1)
        # return result
        return await self.command_exec_t(cmd, log=True, timeout=60)

    async def ifup(self, interface):
        # sleep_sec = 1
        # if interface == PPP_INTERFACE_NAME:
        #     cmd = "pon {0}".format(PPP_PROVIDER)
        #     sleep_sec = GSM_TIMEOUT
        # else:
        #     cmd = "ifup --force {0}".format(interface)

        # result = command_exec(cmd)
        # # if result == 0:
        # time.sleep(sleep_sec)
        # # add_default_dns()
        # return result

        cmd = "ifup --force {0}".format(interface)
        result = await self.command_exec(cmd)
        await asyncio.sleep(DEFAULT_TIMEOUT)
        return result

    async def ifdown(self, interface):
        # sleep_sec = 1
        # if interface == PPP_INTERFACE_NAME:
        #     cmd = "poff {0}".format(PPP_PROVIDER)
        #     sleep_sec = GSM_TIMEOUT
        # else:
        #     cmd = "ifdown --force {0}".format(interface)

        # result = command_exec(cmd)
        # # if result == 0:
        # time.sleep(sleep_sec)
        # # add_default_dns()
        # return result

        cmd = "ifdown --force {0}".format(interface)
        return await self.command_exec(cmd)

    async def update_interface_settings(self, interface_settings):
        shutil.copy(LINUX_INTERFACES_FILE_PATH, LINUX_INTERFACES_BACKUP_FILE_PATH)

        interfaces_text = "\n\n".join([d for (i, d, m) in interface_settings]) + "\n"
        with open(LINUX_INTERFACES_FILE_PATH, 'w') as f:
            f.write(interfaces_text)

        # ifdown -i /etc/network/interfaces.bak <your_interface>; ifup <your_interface>
        await self.command_exec("ifdown -i {0} -a".format(LINUX_INTERFACES_BACKUP_FILE_PATH))
        await self.command_exec("ifup -a")

        # (sleep 180; ifdown <your_interface>; ifup -i /etc/network/interfaces.bak <your_interface>)&

    async def restore_interface_settings(self, interface_settings):
        if not os.path.isfile(LINUX_INTERFACES_BACKUP_FILE_PATH):
            return False

        await self.command_exec("ifdown -a")
        await self.command_exec("ifup -i {0} -a".format(LINUX_INTERFACES_BACKUP_FILE_PATH))
        return True

    @staticmethod
    def _parse_wlan_quality_param(param_str):
        # param_str example: 'Signal level=100/100  '
        tmp = param_str.split('=')
        if len(tmp) > 1:
            return tmp[-1].strip()

    async def wlan_quality(self):
        cmd = "iwconfig wlan0 | grep -i quality"
        # res, out, err = await self.command_exec(cmd)
        try:
            res, out, err = await asyncio.wait_for(self.command_exec(cmd, log=False), timeout=60)
        except asyncio.TimeoutError:
            self.logger.error("wlan_quality: command_exec timeout")
            return

        if res != 0:
            return

        # stdout example: '          Link Quality=98/100  Signal level=100/100  Noise level=0/100'
        out = out.strip().lower()

        params = ["link quality", "signal level", "noise level"]
        tmp = []
        for param in params:
            i = out.find(param)
            if i < 0:
                continue
            tmp.append((i, param))
        if not tmp:
            return

        result = {}
        tmp.sort(key=lambda x: x[0])
        tmp_len = len(tmp)
        if tmp_len > 1:
            for i in range(tmp_len-1):
                result[tmp[i][1]] = self._parse_wlan_quality_param(out[tmp[i][0]:tmp[i+1][0]])
        result[tmp[-1][1]] = self._parse_wlan_quality_param(out[tmp[-1][0]:])
        return result

    async def interface_quality(self, interfaces):
        result = {}
        for interface in interfaces:
            if (interface == PPP_INTERFACE_NAME) or (interface == USB0_INTERFACE_NAME):
                quality = await self.modem_quality()
            elif interface == WLAN_INTERFACE_NAME:
                quality = await self.wlan_quality()
            else:
                continue
            if quality is not None:
                result[interface] = quality
        return result


def skip_interface(interface):
    try:
        if interface in [ETH0_INTERFACE_NAME, ETH1_INTERFACE_NAME]:  # , WLAN_INTERFACE_NAME]:
            # File indicates the current physical link state of the interface.
            filepath = "/sys/class/net/{0}/carrier".format(interface)
            if os.path.isfile(filepath):
                with open(filepath, 'r') as file:
                    state_txt = file.read()
                    try:
                        state = int(state_txt.strip())
                        if state == 0:
                            return True
                    except ValueError:
                        pass
    except Exception:
        pass
    return False

###################################################################################################


def rx_tx_bytes(interfaces=None):
    with open("/proc/net/dev", 'r') as file:
        lines = file.readlines()
    if not lines or len(lines) < 2:
        return {}
    columnLine = lines[1]
    _, receiveCols, transmitCols = columnLine.split("|")
    cols = []
    for pref, tmp_cols in [('recv_', receiveCols.split()), ('trans_', transmitCols.split())]:
        cols += [pref+field for field in tmp_cols]
    result = {}
    for line in lines[2:]:
        if line.find(":") < 0:
            continue
        interface, data = line.split(":")
        interface = interface.strip()
        if (interfaces is not None) and (interface not in interfaces):
            continue
        interface_data = dict(zip(cols, [int(d) for d in data.split()]))
        result[interface] = interface_data
    return result


class BytesCounter:
    def __init__(self):
        self._stat = {}  # interface: (timestamp, recv, trans)
    def count(self, interfaces=None):
        result = {}
        ts = datetime.datetime.now().timestamp()
        tmp = rx_tx_bytes(interfaces)
        for interface, interface_data in tmp.items():
            current_stat = ts, interface_data.get('recv_bytes'), interface_data.get('trans_bytes')
            prev_stat = self._stat.get(interface)
            if prev_stat is None:
                self._stat[interface] = current_stat
                continue
            d = ts - prev_stat[0]
            if d <= 0:
                continue
            if d > 2*60:
                self._stat[interface] = current_stat
                continue
            self._stat[interface] = current_stat
            r = None
            if (current_stat[1] is not None) and (prev_stat[1] is not None):
                r = (current_stat[1]-prev_stat[1]) / d
                if r < 0:
                    r = None
            t = None
            if (current_stat[2] is not None) and (prev_stat[2] is not None):
                t = (current_stat[2]-prev_stat[2]) / d
                if t < 0:
                    t = None
            result[interface] = ts, r, t
        return result
