import os
import sys
import importlib.util

import logging
import logging.handlers

from datetime import datetime

import asyncio


PATH_HEADER = "PATH=$(su - root -c \"echo \\$PATH\")\n"

LOG_FORMAT = ">>> %(asctime)s | %(levelname)8s | [%(module)s]:[%(funcName)s]\n%(message)s"
LOG_TIME_FORMAT = '%d.%m.%Y %H:%M:%S'

LOG_MAX_BYTES = 1024*1024
LOG_BACKUP_COUNT = 5


def parse_file_path(file_path):
    module_path, module_filename = os.path.split(file_path)
    module_name = os.path.splitext(module_filename)[0]
    return module_name, module_path


def get_logger(filepath, name=None):
    new_log = logging.getLogger(name)
    new_log.setLevel(logging.INFO)

    for handler in reversed(new_log.handlers):
        handler.close()
        new_log.removeHandler(handler)

    # handler = logging.handlers.RotatingFileHandler(filepath,
    #                                                maxBytes=LOG_MAX_BYTES,
    #                                                backupCount=LOG_BACKUP_COUNT)
    handler = logging.handlers.TimedRotatingFileHandler(filepath,
                                                        when='h', interval=1, backupCount=24)
    formatter = logging.Formatter(fmt=LOG_FORMAT, datefmt=LOG_TIME_FORMAT)
    handler.setFormatter(formatter)

    new_log.addHandler(handler)
    new_log.addHandler(logging.StreamHandler(sys.stdout))

    new_log.propagate = False
    return new_log


def get_null_logger():
    new_logger = logging.getLogger('null')
    new_logger.setLevel(logging.INFO)

    for handler in reversed(new_logger.handlers):
        handler.close()
        new_logger.removeHandler(handler)

    new_logger.addHandler(logging.NullHandler())
    new_logger.propagate = False
    return new_logger


def import_module(dir_path, file_name):
    """
    Функция, возвращающая python-модуль по его пути
    """
    file_path = os.path.join(dir_path, file_name)
    module_name, ext = os.path.splitext(file_name)

    spec = importlib.util.spec_from_file_location(module_name, file_path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module


class IQLog:
    def __init__(self, logger=None):
        self.logger = logger if logger is not None else get_null_logger()


class ShellCommand(IQLog):
    def __init__(self, logger=None):
        super().__init__(logger)
        # self.logger = logger if logger is not None else get_null_logger()

    async def _proc_result(self, proc, cmd, log):
        stdout, stderr = await proc.communicate()
        log_txt = "[\"{0}\" exited with {1}]".format(cmd, proc.returncode)
        if stdout:
            stdout = stdout.decode("utf-8")
            log_txt += "\n[stdout]\n{0}".format(stdout)
        if stderr:
            stderr = stderr.decode("utf-8")
            log_txt += "\n[stderr]\n{0}".format(stderr)
        if log:
            self.logger.info(log_txt)
        return proc.returncode, stdout, stderr

    async def command_exec(self, cmd, log=True):
        proc = await asyncio.create_subprocess_shell(
            PATH_HEADER + cmd,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE)
        return await self._proc_result(proc, cmd, log)

    async def command_exec_t(self, cmd, log=True, timeout=None):
        if timeout is None:
            return await self.command_exec(cmd, log)

        proc = await asyncio.create_subprocess_shell(
            PATH_HEADER + cmd,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE)

        is_killed = False
        start_dt = datetime.now()
        while proc.returncode is None:
            await asyncio.sleep(1)
            if (datetime.now() - start_dt).total_seconds() > timeout:
                kill_cmd = "/usr/bin/pkill -P {0}".format(proc.pid)
                r, o, e = await self.command_exec(kill_cmd, log)
                self.logger.warning("KILL [{0}], pid = {1}: r={2}\no={3}\ne={4}".format(cmd, proc.pid, r, o, e))
                is_killed = True
                break
        if is_killed:
            return -1, "kill", "kill"
        return await self._proc_result(proc, cmd, log)
