#!/usr/bin/python3

"""
Модуль, реализующий переключение сетевых интерфейсов на контроллере

"""

# import sys
import os
import json
import socket
# import subprocess
import traceback
# import time
# import shutil
# import string

import asyncio

import datetime
# import paho.mqtt.client as mqtt


# from multiprocessing import Process, Manager

from lib.sys_utils import parse_file_path, get_logger, import_module, IQLog
from lib.mqtt_client import MQTTClient
from lib.net_utils import NetUtils  # , rx_tx_bytes
from lib.gsm_modem import GSMModem  #, MODEM_ON, MODEM_OFF  # , PROVIDER_ERROR, PROVIDER_NEW, PROVIDER_OK

from lib.settings import CONFIG_FILE_NAME, BASE_INTERFACES, \
    INTERFACE_SETTINGS_TEMPLATE, LO_INTERFACE_NAME, WLAN_INTERFACE_NAME, FREE_WLAN_INTERFACE_DATA, PPP_INTERFACE_NAME, \
    DEFAULT_INTERFACE_PRIORITY, DEFAULT_SETTINGS, EMPTY_MODEM_SETTINGS
    # LINUX_INTERFACES_FILE_PATH, DEFAULT_MODEM_DEV, USB0_INTERFACE_NAME  # , USB0_INTERFACE_DATA


STORAGE_SETTINGS = 'settings'
STORAGE_INTERFACES = 'interfaces'
STORAGE_PREV_INTERFACES = 'prev_interfaces'
STORAGE_MODEM = 'modem'

DEFAULT_STORAGE_DATA = {
    STORAGE_SETTINGS: None,
    STORAGE_INTERFACES: [],
    STORAGE_MODEM: EMPTY_MODEM_SETTINGS
}


class IQNetworkManager(IQLog):
    def __init__(self, module_path):
        self.name, self.working_dir = parse_file_path(module_path)
        logger, self.tmp_dir = self._configure()
        super().__init__(logger=logger)

        self._storage = None
        self._storage_file_path = os.path.join(self.tmp_dir, 'storage.json')

        _modem = GSMModem(logger=self.logger)  # self._modem
        self._net = NetUtils(_modem, self.logger)

        self._mqtt_client = MQTTClient(self.logger)

        # self._ppp_is_active = False
        self._no_internet_counter = 0
        # self._restore_mode = False

    def _configure(self):
        config = import_module(self.working_dir, CONFIG_FILE_NAME)

        # Update default logger
        if hasattr(config, 'LOG_FILE_PATH'):
            log_file_path = config.LOG_FILE_PATH
        else:
            log_file_path = os.path.join(self.working_dir, '{0}.log'.format(self.name))

        if hasattr(config, 'TPM_DIR_PATH'):
            tmp_dir_path = config.TPM_DIR_PATH
        else:
            tmp_dir_path = os.path.join(self.working_dir, 'tmp')

        log_name, log_dir_path = parse_file_path(log_file_path)
        for dir_path in [log_dir_path, tmp_dir_path]:
            if not os.path.isdir(dir_path):
                os.makedirs(dir_path)

        logger = get_logger(log_file_path, log_name)
        return logger, tmp_dir_path

    def start(self):
        try:
            try:
                lock_socket = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
                lock_socket.bind('\0' + os.path.join(self.working_dir, self.name))
            except socket.error:
                self.logger.warning("Blocked by socket")
                return

            self.logger.info("Start")

            try:
                self._start()
            except:
                self.logger.critical(traceback.format_exc())

            lock_socket.shutdown(socket.SHUT_RDWR)
            lock_socket.close()
        except:
            self.logger.critical(traceback.format_exc())
        finally:
            self.logger.info("Finish")

    def _start(self):
        loop = asyncio.get_event_loop()
        loop.create_task(self._mqtt_client.start())
        loop.create_task(self._engine())
        # loop.create_task(self._rx_tx_stat())
        loop.run_forever()

    # async def _rx_tx_stat(self):
    #     skip_interfaces = [LO_INTERFACE_NAME]
    #     await asyncio.sleep(10)
    #     while True:
    #         result = {}
    #         prev_ts = datetime.datetime.now().timestamp()
    #         prev_data = rx_tx_bytes()
    #         await asyncio.sleep(60)
    #         ts = datetime.datetime.now().timestamp()
    #         data = rx_tx_bytes()
    #         d = ts - prev_ts
    #         for interface, interface_data in data.items():
    #             if interface in skip_interfaces:
    #                 continue
    #             prev_tmp = prev_data.get(interface)
    #             if prev_tmp is None:
    #                 continue
    #             prev_stat = prev_ts, prev_tmp.get('recv_bytes'), prev_tmp.get('trans_bytes')
    #             stat = ts, interface_data.get('recv_bytes'), interface_data.get('trans_bytes')
    #             r = None
    #             if (stat[1] is not None) and (prev_stat[1] is not None):
    #                 r = (stat[1]-prev_stat[1]) / d
    #                 if r < 0:
    #                     r = None
    #             t = None
    #             if (stat[2] is not None) and (prev_stat[2] is not None):
    #                 t = (stat[2]-prev_stat[2]) / d
    #                 if t < 0:
    #                     t = None
    #             result[interface] = "{0}".format(datetime.datetime.fromtimestamp(ts)), \
    #                                 "{0}".format(datetime.datetime.fromtimestamp(prev_ts)), d, \
    #                                 r, t
    #         if result:
    #             self._mqtt_client.publish_stat(result)

    async def _engine(self):
        await self._initialization()
        await asyncio.sleep(15)

        while True:
            try:
                self.logger.info(">>> ENGINE TASK")
                await self._engine_task()
                self.logger.info("<<< ENGINE TASK\n\n\n")
            except:
                self.logger.critical(traceback.format_exc())
            await asyncio.sleep(60)

    async def _initialization(self):
        await self._net.print_routes()
        self._storage_init()
        self.logger.info("STORAGE init state:\n{0}".format(self._storage_txt()))
        await self._net.modem_on()

    def _storage_init(self):
        file_storage = self._get_storage_from_file()
        if file_storage is None:
            self._storage = DEFAULT_STORAGE_DATA
            self._set_storage_file()
        else:
            self._storage = file_storage

    def _get_storage_from_file(self):
        if os.path.isfile(self._storage_file_path):
            try:
                with open(self._storage_file_path) as f:
                    json_storage = json.load(f)
                return json_storage
            except:
                self.logger.error("Can't read from the STORAGE json-file [{0}]:\n{1}".format(self._storage_file_path,
                                                                                             traceback.format_exc()))
        else:
            self.logger.warning("The STORAGE json-file [{0}] is not file".format(self._storage_file_path))
        return None

    def _set_storage_file(self):
        if self._storage is None:
            os.remove(self._storage_file_path)
        else:
            try:
                with open(self._storage_file_path, 'w+') as f:
                    json.dump(self._storage, f)
            except:
                self.logger.error("Can't write STORAGE json-file [{0}]:\n{1}".format(self._storage_file_path,
                                                                                     traceback.format_exc()))

    def _storage_txt(self):
        return json.dumps(self._storage, indent=4, sort_keys=True) if self._storage is not None else None

    # @staticmethod
    # def _compare_interface_settings(interface_settings1, interface_settings2):
    #     try:
    #         interface_set1 = set()
    #         interface_set2 = set()
    #         for (i_settings, i_set) in [(interface_settings1, interface_set1), (interface_settings2, interface_set2)]:
    #             for (i, d, m) in i_settings:
    #                 i_set.add((i, d))
    #         if interface_set1 == interface_set2:
    #             return True
    #     except:
    #         pass
    #     return False

    @staticmethod
    def _settings_are_changed(settings1, settings2):
        if (settings1 is not None) and (settings2 is not None):
            return settings1 != settings2
        if (settings1 is None) and (settings2 is None):
            return False
        return True

    @staticmethod
    def _interface_data_is_changed(interface_settings1, interface_settings2):
        try:
            interface_set1 = set()
            interface_set2 = set()
            for (i_settings, i_set) in [(interface_settings1, interface_set1), (interface_settings2, interface_set2)]:
                for (i, d, m) in i_settings:
                    i_set.add((i, d))
            if interface_set1 == interface_set2:
                return False
        except:
            pass
        return True

    def _check_for_updates(self):
        interface_data_is_changed = False
        settings_are_changed = False
        client_settings = self._mqtt_client.get_settings()

        if self._storage[STORAGE_SETTINGS] is None:
            # First check !
            if client_settings is None:
                client_settings = DEFAULT_SETTINGS
            settings_are_changed = True
        else:
            if client_settings is not None:
                settings_are_changed = self._settings_are_changed(self._storage[STORAGE_SETTINGS], client_settings)

        if settings_are_changed:
            self.logger.info("Interface settings are changed:\n   new: {0}\n   prev: {1}".format(
                client_settings, self._storage[STORAGE_SETTINGS]))
            self._storage[STORAGE_SETTINGS] = client_settings

            interfaces = self._process_interface_settings()
            interface_data_is_changed = self._interface_data_is_changed(self._storage[STORAGE_INTERFACES], interfaces)
            # if interface_data_is_changed:
            self._storage[STORAGE_PREV_INTERFACES] = self._storage[STORAGE_INTERFACES]
            self._storage[STORAGE_INTERFACES] = interfaces
        return settings_are_changed, interface_data_is_changed

    def _process_interface_settings(self):
        interface_settings = {i: INTERFACE_SETTINGS_TEMPLATE[i][0] for i in BASE_INTERFACES}  # i_name -> i_data(txt)
        interface_priorities = {}

        for interface_name, client_interface_settings in self._storage[STORAGE_SETTINGS].items():
            init_interface_settings = INTERFACE_SETTINGS_TEMPLATE.get(interface_name)
            if init_interface_settings is None:
                continue
            is_active = client_interface_settings.get('is_active', True)

            client_interface_data = client_interface_settings.get('data')
            if (not is_active) or ((client_interface_data is not None) and (not client_interface_data)):
                # if not active or empty string interface_data
                interface_settings.pop(interface_name, None)
                continue

            if client_interface_data is not None:
                interface_data = client_interface_data
            else:
                interface_data = init_interface_settings[0]
                if interface_name == WLAN_INTERFACE_NAME:
                    ssid = client_interface_settings.get('ssid')
                    if (ssid is None) or (not ssid):
                        self.logger.error("There is no \"ssid\" WLAN in the settings json-file")
                        continue
                    key = client_interface_settings.get('key')
                    if (key is not None) and key:
                        interface_data = interface_data.format(ssid, key)
                    else:
                        interface_data = FREE_WLAN_INTERFACE_DATA.format(ssid)
                        self.logger.warning("There is no \"key\" WLAN")

            if interface_data is not None:
                interface_settings[interface_name] = interface_data
                interface_priorities[interface_name] = client_interface_settings.get('priority',
                                                                                     DEFAULT_INTERFACE_PRIORITY)
        # self._ppp_is_active = PPP_INTERFACE_NAME in interface_settings
        # for i in list(interface_settings):
        #     if i == PPP_INTERFACE_NAME:
        #         self._ppp_is_active = True
        #         if self._net.is_4g_modem():
        #             interface_settings[USB0_INTERFACE_NAME] = INTERFACE_SETTINGS_TEMPLATE[USB0_INTERFACE_NAME][0]
        #             interface_priorities[USB0_INTERFACE_NAME] = interface_priorities.get(i, DEFAULT_INTERFACE_PRIORITY)
        #             interface_settings.pop(i, None)
        #             interface_priorities.pop(i, None)

        prioritized_interfaces = [(i, d, interface_priorities.get(i, DEFAULT_INTERFACE_PRIORITY),
                                   INTERFACE_SETTINGS_TEMPLATE[i][1]) for i, d in interface_settings.items()]
        prioritized_interfaces.sort(key=lambda x: (x[2], x[3]))  # sort by pair (PRIORITY, ORDER)

        result_interfaces_data = []
        metric = 0
        for (i, d, p, o) in prioritized_interfaces:
            if i == LO_INTERFACE_NAME:
                result_interfaces_data.insert(0, [i, d, 0])
                continue
            result_interfaces_data.append([i, d, metric])
            metric += 10
        return result_interfaces_data

    async def _check_interface(self, interface, current_metrics, target_metric, checked_interfaces):
        check_result = await self._net.check_interface(interface, current_metrics, target_metric)
        if check_result:
            checked_interfaces.append(interface)

    def _ppp_is_active(self):
        try:
            for i, d, m in self._storage[STORAGE_INTERFACES]:
                if i == PPP_INTERFACE_NAME:
                    return True
        except:
            return True
        return False

    async def _check_modem_settings(self):
        if not self._ppp_is_active():
            return
        await self._net.check_modem()
        tmp_modem_settings = self._net.modem_settings()
        if self._settings_are_changed(self._storage[STORAGE_MODEM], tmp_modem_settings):
            self._storage[STORAGE_MODEM] = tmp_modem_settings
            self._set_storage_file()
            self.logger.info("STORAGE [updated by modem]:\n{0}".format(self._storage_txt()))

    async def _engine_task(self):
        self._net.check_dns()

        if await self._net.internet_via_eth():  # For controller service needs
            self._no_internet_counter = 0
            self.logger.info("(!) ETHERNET CONNECTION, skip network settings checking")
            return

        settings_are_changed, interface_data_is_changed = self._check_for_updates()
        if settings_are_changed:
            self._set_storage_file()
            self.logger.info("(!!!) STORAGE [updated by settings]:\n{0}".format(self._storage_txt()))

            self.logger.warning("(!!!) START [UPDATE INTERFACES BY NEW SETTINGS]")
            if interface_data_is_changed:
                self.logger.warning(">>> (!!!) UPDATE INTERFACE DATA")
                await self._net.update_interface_settings(self._storage[STORAGE_INTERFACES])
            self.logger.info(">>> (!!!) DELETE ALL DEFAULT GATEWAYS")
            await self._net.delete_default_gateways()
            self.logger.warning("(!!!) FINISH [UPDATE INTERFACES BY NEW SETTINGS]")

        if not self._storage[STORAGE_INTERFACES]:
            self.logger.info("There is no any actual INTERFACES... Skip this settings")
            return

        await self._check_modem_settings()

        current_gateways = self._net.get_current_gateways()
        self.logger.info("current_gateways = {0}".format(current_gateways))
        await self._net.print_routes()

        target_interfaces = {}
        checked_interfaces = []
        # check_tasks = []
        for (i, d, m) in self._storage[STORAGE_INTERFACES]:
            if i == LO_INTERFACE_NAME:
                continue

            # iface = PPP_INTERFACE_NAME if self._net.is_4g_modem() and (i == USB0_INTERFACE_NAME) else i
            iface = i

            target_interfaces[iface] = m
            self.logger.info("CHECK [{0}], metric = {1}".format(i, m))
            current_metrics = current_gateways.get(i, [])

            check_result = await self._net.check_interface(i, current_metrics, m)
            if check_result:
                checked_interfaces.append(iface)
            # check_tasks.append(self._check_interface(i, current_metrics, m, checked_interfaces))

        # await asyncio.gather(*check_tasks)
        # self.logger.info("CHECKED interfaces {0}".format(checked_interfaces))
        await self._net.print_routes()

        # actual_interfaces = {}
        # for i, m in self._net.get_current_gateways().items():
        #     if self._net.is_4g_modem() and (i == USB0_INTERFACE_NAME):
        #         i = PPP_INTERFACE_NAME
        #     actual_interfaces[i] = m
        actual_interfaces = self._net.get_current_gateways()

        current_datetime = datetime.datetime.now()
        result = {
            'event_timestamp': current_datetime.timestamp(),
            'datetime': str(current_datetime),  # TEMP FIELD
            'current_interfaces': {checked_interfaces[i]: i for i in range(len(checked_interfaces))},  # TEMP FIELD
            'target_interfaces': target_interfaces,
            'actual_interfaces': actual_interfaces,
            'checked_interfaces': checked_interfaces,
            'interface_quality': await self._net.interface_quality(checked_interfaces)
        }

        self.logger.info("Start publish result:\n{0}".format(result))
        r = self._mqtt_client.publish_result(result)
        self.logger.info("Publish result: {0}".format(r))
        # self.logger.info("Publish result [status - {0}]:\n{1}".format(r, result))
        # mqtt_client.publish(LOG_MQTT_TOPIC, LOG_RESULT_JSON, qos=2, retain=False)

        if not checked_interfaces:
            self._no_internet_counter += 1
            self.logger.error("!!! NO INTERNET: {0}".format(self._no_internet_counter))
            # if self._no_internet_counter > 5:
            #     if self._net.restore_interface_settings():
            #         self._restore_mode = True

            self.logger.info("Sleep 1 min")
            await asyncio.sleep(60)
        else:
            self._no_internet_counter = 0


if __name__ == "__main__":
    nm = IQNetworkManager(os.path.realpath(__file__))
    nm.start()
